float rgb_2_saturation( vec3 rgb )
{
	float minrgb = min( min(rgb.r, rgb.g ), rgb.b );
	float maxrgb = max( max(rgb.r, rgb.g ), rgb.b );
	return ( max( maxrgb, 1e-10 ) - max( minrgb, 1e-10 ) ) / max( maxrgb, 1e-2 );
}

float glow_fwd( float yc_in, float glow_gain_in, float glow_mid)
{
   float glow_gain_out;

   if (yc_in <= 2./3. * glow_mid) {
	 glow_gain_out = glow_gain_in;
   } else if ( yc_in >= 2 * glow_mid) {
	 glow_gain_out = 0;
   } else {
	 glow_gain_out = glow_gain_in * (glow_mid / yc_in - 0.5);
   }

   return glow_gain_out;
}

float glow_inv( float ycOut, float glow_gain_in, float glow_mid)
{
	float glow_gain_out;

	if (ycOut <= ((1 + glow_gain_in) * 2./3. * glow_mid)) {
		glow_gain_out = -glow_gain_in / (1 + glow_gain_in);
	} else if ( ycOut >= (2. * glow_mid)) {
		glow_gain_out = 0.;
	} else {
		glow_gain_out = glow_gain_in * (glow_mid / ycOut - 1./2.) / (glow_gain_in / 2. - 1.);
	}

	return glow_gain_out;
}


float sigmoid_shaper( float x)
{
	float t = max( 1 - abs( 0.5 * x ), 0 );
	float y = 1 + sign(x) * (1 - t*t);
	return 0.5 * y;
}

float rgb_2_hue( vec3 rgb ) 
{
	float hue;
	if (rgb[0] == rgb[1] && rgb[1] == rgb[2])
	{
		hue = 0;
	}
	else
	{
		hue = (180.0 / pi) * atan(2.0 * rgb[0] - rgb[1] - rgb[2], sqrt(3.0) * (rgb[1] - rgb[2]));
	}

	if (hue < 0.)
		hue = hue + 360;

	return clamp( hue, 0, 360 );
}

float rgb_2_yc( vec3 rgb, float ycRadiusWeight)
{
	float r = rgb[0]; 
	float g = rgb[1]; 
	float b = rgb[2];
  
	float chroma = sqrt(b*(b-g)+g*(g-r)+r*(r-b));

	return ( b + g + r + ycRadiusWeight * chroma) / 3.;
}

float rgb_2_yc( vec3 rgb)
{
return rgb_2_yc(rgb, 1.75);
}

float slope = ACES_Slope;
float toe = ACES_Toe;
float shoulder = ACES_Shoulder;
float black_clip = ACES_BlackClip;
float white_clip = ACES_WhiteClip;
float expand_gamut = ACES_GamutExpansion;

vec3 film_tonemap(vec3 in_color) 
{
	mat3 AP0_2_sRGB = (AP0_2_XYZ_MAT * D60_2_D65_CAT) * XYZ_2_sRGB_MAT;
	mat3 AP1_2_sRGB = (AP1_2_XYZ_MAT * D60_2_D65_CAT) * XYZ_2_sRGB_MAT;
	
	mat3 AP0_2_AP1 = AP0_2_XYZ_MAT * XYZ_2_AP1_MAT;
	mat3 AP1_2_AP0 = AP1_2_XYZ_MAT * XYZ_2_AP0_MAT;

    vec3 color_AP1 = (in_color * sRGB_2_AP0) * AP0_2_AP1;
    float luma_AP1 = dot(color_AP1, AP1_RGB2Y);

    vec3 chroma_AP1 = color_AP1 / luma_AP1;

    float chroma_dist_sqr = dot(chroma_AP1 - 1.0, chroma_AP1 - 1);
    float expand_amount = (1.0 - exp2(-4.0 * chroma_dist_sqr)) * (1.0 - exp2(-4.0 * expand_gamut * luma_AP1 * luma_AP1));

    const mat3 Wide_2_XYZ_MAT = mat3(
        vec3(0.5441691,  0.2395926,  0.1666943),
        vec3(0.2394656,  0.7021530,  0.0583814),
        vec3(-0.0023439,  0.0361834,  1.0552183)
    );

    const mat3 Wide_2_AP1 = Wide_2_XYZ_MAT * XYZ_2_AP1_MAT;
    const mat3 expand_mat = AP1_2_sRGB * Wide_2_AP1;

    vec3 color_expand = color_AP1 * expand_mat;
    color_AP1 = mix(color_AP1, color_expand, expand_amount);

	vec3 color_AP0 = color_AP1 * AP1_2_AP0;

	// "Glow" module constants
	float rrt_glow_gain = 0.05;
	float rrt_glow_mid = 0.08;

	float saturation = rgb_2_saturation(color_AP0);
	float yc_in = rgb_2_yc(color_AP0);
	float s = sigmoid_shaper((saturation - 0.4) * 5.0);
	float added_glow = 1 + glow_fwd(yc_in, rrt_glow_gain * s, rrt_glow_mid);
	color_AP0 *= added_glow;

	// Use ACEScg primaries as working space
	vec3 working_color = color_AP0 * AP0_2_AP1_MAT;
	working_color = max(vec3(0.0), working_color);
	
	float toe_scale		= 1 + black_clip - toe;
	float shoulder_scale	= 1 + white_clip - shoulder;
	
	float in_match = 0.18;
	float out_match = 0.18;

	float toe_match = 0.0;
	if(toe > 0.8)
	{
		// 0.18 will be on straight segment
		toe_match = (1 - toe  - out_match) / slope + log10(in_match);
	}
	else
	{
		// 0.18 will be on toe segment

		// Solve for toe_match such that input of in_match gives output of out_match.
		float bt = (out_match + black_clip) / toe_scale - 1;
		toe_match = log10(in_match) - 0.5 * log((1+bt)/(1-bt)) * (toe_scale / slope);
	}

	float straight_match = (1 - toe) / slope - toe_match;
	float shoulder_match = shoulder / slope - straight_match;
	
	vec3 log_color = log10(working_color);
	vec3 straight_color = slope * (log_color + straight_match);
	
	vec3 toe_color		= (-black_clip) + (2 * toe_scale) / (1 + exp((-2 * slope / toe_scale) * (log_color - toe_match)));
	vec3 shoulder_color	= (1 + white_clip) - (2 * shoulder_scale) / (1 + exp(( 2 * slope / shoulder_scale) * (log_color - shoulder_match)));

    toe_color.x = log_color.x < toe_match ? toe_color.x : straight_color.x;
    toe_color.y = log_color.y < toe_match ? toe_color.y : straight_color.y;
    toe_color.z = log_color.z < toe_match ? toe_color.z : straight_color.z;

    shoulder_color.x = log_color.x > shoulder_match ? shoulder_color.x : straight_color.x;
    shoulder_color.y = log_color.y > shoulder_match ? shoulder_color.y : straight_color.y;
    shoulder_color.z = log_color.z > shoulder_match ? shoulder_color.z : straight_color.z;

	vec3 t = saturate((log_color - toe_match) / (shoulder_match - toe_match));
	t = shoulder_match < toe_match ? 1 - t : t;
	t = (3-2*t)*t*t;
	vec3 tone_color = mix(toe_color, shoulder_color, t);

	tone_color = max(tone_color, 0.0);

	// Returning positive AP1 values
	return max(vec3(0), tone_color * AP1_2_sRGB);
}

vec3 tonemap(in vec3 x) { return film_tonemap(x); }