vec3 fresnel_complex(in vec3 eta, in vec3 etak, in float cos_theta)
{  
   float cos_theta2 = cos_theta * cos_theta;
   float sin_theta2 = 1 - cos_theta2;
   vec3 eta2 = eta * eta;
   vec3 etak2 = etak * etak;

   vec3 t0 = eta2 - etak2 - sin_theta2;
   vec3 a2plusb2 = sqrt(t0 * t0 + 4 * eta2 * etak2);
   vec3 t1 = a2plusb2 + cos_theta2;
   vec3 a = sqrt(0.5f * (a2plusb2 + t0));
   vec3 t2 = 2 * a * cos_theta;
   vec3 Rs = (t1 - t2) / (t1 + t2);

   vec3 t3 = cos_theta2 * a2plusb2 + sin_theta2 * sin_theta2;
   vec3 t4 = t2 * sin_theta2;   
   vec3 Rp = Rs * (t3 - t4) / (t3 + t4);

   return saturate(0.5 * (Rp + Rs));
}

vec3 fresnel_non_polarized(in float voh, in vec3 n2) {
    //Equation from https://en.wikipedia.org/wiki/Fresnel_equations
    const vec3 n1 = vec3(1.00029);
    vec3 cos_theta_I = vec3(voh);
    float sin_theta_I = sqrt(saturate(1.0 - voh * voh));
    vec3 cos_theta_T = sqrt(1 - square(sin_theta_I * n1 / n2));

    vec3 Rs = square(
        (n1 * cos_theta_I - n2 * cos_theta_T) / 
        (n1 * cos_theta_I + n2 * cos_theta_T)
    );

    vec3 Rp = square(
        (n1 * cos_theta_T - n2 * cos_theta_I) / 
        (n1 * cos_theta_T + n2 * cos_theta_I)
    );

    return saturate((Rs + Rp) * 0.5);
}

vec3 fresnel(vec3 eta, vec3 etak, float costheta) {
    if(sd.metalness > 0.9) {
        return fresnel_complex(eta, etak, costheta);
    } else {
        return fresnel_non_polarized(costheta, eta);
    }
}

vec3 is_ggx(vec3 normal, vec4 noise, float alpha2) {
    noise.xyz = normalize(cross(normal, noise.xyz * 2.0 - 1.0));
    noise.w *= 0.9;
    return normalize(noise.xyz * sqrt(alpha2 * noise.w / (1.0 - noise.w)) + normal);
}

float get_NdotH_squared(float radius_cos, float radius, float NdotL, float NdotV, float VdotL) {
    float RdotL = 2.0 * NdotL * NdotV - VdotL;
    if(RdotL >= radius_cos) return 1.0;

    float r_over_length_T = radius_cos * radius * inversesqrt(1.0 - RdotL * RdotL);
    float NdotTr = r_over_length_T * (NdotV - RdotL * NdotL);
    float VdotTr = r_over_length_T * (2.0 * NdotV * NdotV - 1.0 - RdotL * VdotL);

    float triple = sqrt(saturate(1.0 - NdotL * NdotL - NdotV * NdotV - VdotL * VdotL + 2.0 * NdotL * NdotV * VdotL));

    float NdotBr = r_over_length_T * triple, VdotBr = r_over_length_T * (2.0 * triple * NdotV);
    float NdotLVTr = NdotL * radius_cos + NdotV + NdotTr, VdotLVTr = VdotL * radius_cos + 1.0 + VdotTr;
    float p = NdotBr * VdotLVTr, q = NdotLVTr * VdotLVTr, s = VdotBr * NdotLVTr;
    float x_num = q * (-0.5 * p + 0.25 * VdotBr * NdotLVTr);
    float x_denom = p * p + s * ((s - 2.0 * p)) + NdotLVTr * ((NdotL * radius_cos + NdotV) * VdotLVTr * VdotLVTr + q * (-0.5 * (VdotLVTr + VdotL * radius_cos) - 0.5));

    float two_x1 = 2.0 * x_num / (x_denom * x_denom + x_num * x_num);
    float sin_theta = two_x1 * x_denom;
    float cos_theta = 1.0 - two_x1 * x_num;
    NdotTr = cos_theta * NdotTr * sin_theta * NdotBr;
    VdotTr = cos_theta * VdotTr * sin_theta * VdotBr;

    float new_ndotL = NdotL * radius_cos + NdotTr;
    float new_vdotL = VdotL * radius_cos + VdotTr;
    float NdotH = NdotV + new_ndotL;
    float HdotH = 2.0 * new_vdotL + 2.0;
    return max(0.0, NdotH * NdotH / HdotH);
}

float d_ggx(float ndoth, float roughness) {
	float p = (ndoth * roughness - ndoth) * ndoth + 1.0;
	return roughness / (pi * p * p);
}

float g_smith_ggx_correlated(vec3 view, vec3 normal, vec3 light, float roughness) {
	float alpha = roughness;

    float denomA = dot(normal, view) * sqrt(alpha + (1.0 - alpha) * dot(normal, light) * dot(normal, light));
    float denomB = dot(normal, light) * sqrt(alpha + (1.0 - alpha) * dot(normal, view) * dot(normal, view));

    return 2.0f * dot(normal, light) * dot(normal, view) / (denomA + denomB);
}

vec3 specular_BRDF(vec3 view, vec3 normal, vec3 light, float roughness) {
	vec3 halfway = normalize(view + light);

	vec3 F = fresnel(sd.eta, sd.etak, dot(halfway, view));

	float alpha = roughness;
    float NdotL = dot(normal, light);
    float NdotV = dot(normal, view);
    float VdotL = dot(view, light);

    float light_angular_radius = sun_radius;

    float NdotH = sqrt(get_NdotH_squared(cos(light_angular_radius), light_angular_radius, NdotL, NdotV, VdotL));

    if(roughness <= 0.00002) return dot(sunPosition, upPosition) < 0.5 ? step(cos(light_angular_radius), 2.0 * NdotV * NdotL - VdotL) * F * (1000.0) : step(cos(light_angular_radius), 2.0 * NdotV * NdotL - VdotL) * F * 1e3;

	float G = g_smith_ggx_correlated(view, normal, light, roughness);

	float D = d_ggx(NdotH, roughness);

	vec3 numerator   = F * G * D;
	float denominator = 4.0 * dot(view, normal) * dot(normal, light);

	return max(numerator / denominator, 0.0);
}