layout(location = 0) out vec3 color;
layout(location = 1) out vec4 transparent;
layout(location = 2) out vec4 albedo;

#include "/lib/important.glsl"
#include "/lib/color/color_space.glsl"
#include "/lib/encoders/encoder.glsl"

//Samplers
uniform sampler2D colortex0;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex5;
uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;
uniform sampler2D noisetex;

//Uniforms
uniform mat4 gbufferModelViewInverse, gbufferProjectionInverse;
uniform mat4 gbufferModelView, gbufferProjection;
uniform mat4 shadowModelView, shadowProjection;

uniform vec3 sun_vector, moon_vector, up_vector, shadow_light_vector, sunPosition, upPosition, cameraPosition;

uniform vec2 view_size, view_pixel_size, taa_offset;

uniform float frameTimeCounter, rainStrength, eyeAltitude;

uniform int isEyeInWater, frameCounter, frame255;

//Fragment inputs
struct light_col {
    vec3 direct;
    vec3 sky;
};

flat in vec4[3] shcoeffs;
flat in light_col light;
in vec2 texture_coordinate;

#include "/lib/converters/converters.glsl"
#include "/lib/structs/posdepth_struct.glsl"
#include "/lib/structs/surface_struct.glsl"
#include "/lib/sky/atmos_projection.glsl"
#include "/lib/color/blackbody_functions.glsl"
#include "/lib/color/light_colors.glsl"
#include "/lib/fragment/water_constants.glsl"
#include "/lib/sky/atmosphere/constants.glsl"
#include "/lib/sky/atmosphere/phase.glsl"
#include "/lib/sky/clouds_2d.glsl"
//#include "/lib/sky/clouds.glsl"
#include "/lib/fragment/shading/apply.glsl"
#include "/lib/sky/atmosphere/transmittance.glsl"

vec3 physical_sun(in vec3 view_vector) {
    float size = sun_radius;
    float cos_theta = dot(view_vector, sun_vector);
    float center_to_edge = clamp(acos(cos_theta) / size, 0.0, 1.0);
    float cos_sun_radius = cos(size);
    if(cos_theta < cos_sun_radius) return vec3(0.0);

    center_to_edge = center_to_edge;
    float mu = sqrt(1.0 - center_to_edge * center_to_edge);

    vec3 a0 = vec3(0.34685, 0.26073, 0.15248);
    vec3 a1 = vec3(1.37539, 1.27428, 1.38517);
    vec3 a2 = vec3(-2.04425, -1.30352, -1.49615);
    vec3 a3 = vec3(2.70493, 1.47085, 1.99886);
    vec3 a4 = vec3(-1.94290, -0.96618, -1.48155);
    vec3 a5 = vec3(0.55999, 0.26384, 0.44119);

    float mu2 = mu * mu;
    float mu3 = mu2 * mu;
    float mu4 = mu2 * mu2;
    float mu5 = mu4 * mu;

    vec3 a05 = (a0 + a1 + a2 + a3 + a4 + a5) * 0.5;

    vec3 sun_spot = exp2(log2(1.0 - center_to_edge * center_to_edge) * a05);

    vec3 sun = sun_color * sun_spot;
    sun *= 1e3;

    vec3 factor = a0 + a1 * mu + a2 * mu2 + a3 * mu3 + a4 * mu4 + a5 * mu5;

	return sun * max(factor, 0.0);
}

vec3 calculate_cloud_fade(mat3 atten_coeff) {
    float cloud_distance_planet = fLength(fNormalize(pd.scene_position)) * fNormalize(pd.scene_position).y * 0.5 + eyeAltitude + planet_radius;
    vec2 rm = exp(-cloud_distance_planet * inverse_scale_heights + scaled_planet_radius);
    float ozone = exp(-max(0.0, (35000.0 - cloud_distance_planet) - atmosphere_radius) /  5000.0)
	            * exp(-max(0.0, (cloud_distance_planet - 35000.0) - atmosphere_radius) / 15000.0);
    vec3 rmo = vec3(rm, ozone);

    return atten_coeff * rmo;
}

#include "/lib/fragment/brdf.glsl"

/* DRAWBUFFERS:01 */
void main() {
    pd = fill_pd_struct(texture_coordinate);
    sd = fill_sd_struct(texture_coordinate);

    color = srgb_to_linear(texture(colortex0, texture_coordinate).rgb);

    bool is_sky = land_mask(pd.depth0) ? false : true;

    if(!is_sky) {
        color = apply_shading(color, true) * (1.0 - sd.metalness);
        #ifdef SunReflection
            if(sd.f0 > 0.0) {
                float ndotl = dot(mat3(gbufferModelView) * decodeNormal(decode2x8(texture(colortex5, texture_coordinate).r)), mat3(gbufferModelView) * shadow_light_vector);

                if(ndotl < 0.0) {
                    vec3 shadows = vec3(0.0);
                } else {
                    vec3 shadows = calculate_shadow_map(pd.shadow_clip_position, false);
                    shadows *= max(0.0, ndotl);
                    shadows *= light.direct;
                    shadows *= cloud_shadow_2d(shadow_light_vector, pd.scene_position + cameraPosition);

                    color += dot(sunPosition, upPosition) < 0.5 ? vec3(0.0) : min(specular_BRDF(-normalize(pd.view_position), sd.vnormal, mat3(gbufferModelView) * sun_vector, sd.roughness_squared), 16e3*16e3) * shadows;
                    color += dot(sunPosition, upPosition) < 0.5 ? min(specular_BRDF(-normalize(pd.view_position), sd.vnormal, mat3(gbufferModelView) * moon_vector, sd.roughness_squared), 16e3*16e3) * shadows : vec3(0.0);
                }
            }
        #endif
    } else {
        #if World == 0
            float dither = fract(frameCounter * (1.0 / 7.0) + bayer32(gl_FragCoord.st));
            //vec4 clouds = clouds_calculate(vec3(0.0), pd.view_position, is_sky, dither, 32.0, 8.0);
            vec4 clouds = clouds(fNormalize(pd.scene_position));

            vec3 cloud_mix = calculate_cloud_fade(coefficients_extinction);
        #elif World == 1
            vec4 clouds = vec4(0.0, 0.0, 0.0, 1.0);
            vec3 cloud_mix = vec3(0.0);
        #endif

        #ifdef Atmos_Analytical
            vec3 sun_transmittance = absorb_analytical(fNormalize(pd.scene_position), coefficients_analytical);
        #else
            vec3 sun_transmittance = atmosphere_transmittance(fNormalize(pd.scene_position), vec3(0.0, planet_radius + eyeAltitude, 0.0), coefficients_extinction);
        #endif

        vec3 sun = physical_sun(fNormalize(pd.scene_position)) * sun_transmittance;
        sun *= clouds.a;
        sun = mix(sun, vec3(0.0), rainStrength);

        color = mix(texture(colortex3, project_sky(fNormalize(pd.scene_position))).rgb, sun * sun_transmittance + texture(colortex3, project_sky(fNormalize(pd.scene_position))).rgb * clouds.a + clouds.rgb, exp(-calculate_cloud_fade(coefficients_extinction) * cloud_distance_2d));
    }

    color /= MaxColorRange;

    transparent = vec4(0.0);
}